/*
 * Copyright (C) 2010 Google Inc. All rights reserved.
 * Copyright (c) 2012 NINTENDO.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above
 * copyright notice, this list of conditions and the following disclaimer
 * in the documentation and/or other materials provided with the
 * distribution.
 *     * Neither the name of Google Inc. nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


#ifndef LayerWiiU_h
#define LayerWiiU_h

#if USE(ACCELERATED_COMPOSITING)

#include "FloatPoint.h"
#include "FloatQuad.h"
#include "GraphicsContext.h"
#include "GraphicsLayerWiiU.h"
#include "PlatformString.h"
#include "TransformationMatrix.h"

#include <wtf/HashMap.h>
#include <wtf/OwnPtr.h>
#include <wtf/PassRefPtr.h>
#include <wtf/RefCounted.h>
#include <wtf/Vector.h>

#include <wtf/text/StringHash.h>
#include <wtf/text/WTFString.h>
#include <wtf/text/CString.h>

namespace WebCore {

class LayerSorterWiiU {
    WTF_MAKE_NONCOPYABLE(LayerSorterWiiU);
public:
    LayerSorterWiiU() : m_zRange(0) { }

    typedef Vector<RefPtr<LayerWiiU> > LayerList;

    void addLayer(RefPtr<LayerWiiU> layer);
    void sort();


    struct LayerShape {
        LayerShape() { }
        LayerShape(float width, float height, const TransformationMatrix& drawTransform);

        float layerZFromProjectedPoint(const FloatPoint&) const;

        FloatPoint3D layerNormal;
        FloatPoint3D transformOrigin;
        FloatQuad projectedQuad;
        FloatRect projectedBounds;
    };

    enum ABCompareResult {
        ABeforeB,
        BBeforeA,
        None
    };

    static ABCompareResult checkOverlap(LayerShape*, LayerShape*, float zThreshold, float& weight);

    LayerList m_sortedList;
private:
    struct GraphEdge;

    struct GraphNode {
        explicit GraphNode(LayerWiiU* wiiulayer) : layer(wiiulayer), incomingEdgeWeight(0) { }
        LayerWiiU* layer;
        LayerShape shape;
        Vector<GraphEdge*> incoming;
        Vector<GraphEdge*> outgoing;
        float incomingEdgeWeight;
    };

    struct GraphEdge {
        GraphEdge(GraphNode* fromNode, GraphNode* toNode , float weight) : from(fromNode), to(toNode) , weight(weight) { };

        GraphNode* from;
        GraphNode* to;
        float weight;
    };

    typedef Vector<GraphNode> NodeList;
    typedef Vector<GraphEdge> EdgeList;
    NodeList m_nodes;
    EdgeList m_edges;
    float m_zRange;

    typedef HashMap<GraphEdge*, GraphEdge*> EdgeMap;
    EdgeMap m_activeEdges;

    void sort(LayerList::iterator first, LayerList::iterator last);
    void createGraphNodes(LayerList::iterator first, LayerList::iterator last);
    void createGraphEdges();
    void removeEdgeFromList(GraphEdge*, Vector<GraphEdge*>&);

};


// Base class for composited layers. Special layer types are derived from
// this class.
class LayerWiiU : public RefCounted<LayerWiiU> {

public:
    static PassRefPtr<LayerWiiU> create(GraphicsLayerWiiU* owner = 0);

    virtual ~LayerWiiU();

    const LayerWiiU* rootLayer() const;
    LayerWiiU* parent() const;
    void addChild(PassRefPtr<LayerWiiU>);
    void insertChild(PassRefPtr<LayerWiiU>, size_t index);
    void replaceChild(LayerWiiU* reference, PassRefPtr<LayerWiiU> newLayer);
    void removeFromParent();
    void removeAllChildren();
    void setChildren(const Vector<RefPtr<LayerWiiU> >&);
    const Vector<RefPtr<LayerWiiU> >& children() const { return m_children; }

    void setAnchorPoint(const FloatPoint& anchorPoint) { m_anchorPoint = anchorPoint; setNeedsCommit(); }
    FloatPoint anchorPoint() const { return m_anchorPoint; }

    void setAnchorPointZ(float anchorPointZ) { m_anchorPointZ = anchorPointZ; setNeedsCommit(); }
    float anchorPointZ() const { return m_anchorPointZ; }

    void setBackgroundColor(const Color& color) { m_backgroundColor = color; setNeedsCommit(); }
    Color backgroundColor() const { return m_backgroundColor; }

    void setBounds(const IntSize&);
    const IntSize& bounds() const { return m_bounds; }
    virtual IntSize contentBounds() const { return bounds(); }

    void setMasksToBounds(bool masksToBounds) { m_masksToBounds = masksToBounds; }
    bool masksToBounds() const { return m_masksToBounds; }

    void setName(const String&);
    const String& name() const { return m_name; }

    void setMaskLayer(LayerWiiU* maskLayer) { m_maskLayer = maskLayer; }
    LayerWiiU* maskLayer() const { return m_maskLayer.get(); }

    void setNeedsDisplay(const FloatRect& dirtyRect);
    void setNeedsDisplay();
    const FloatRect& dirtyRect() const { return m_dirtyRect; }
    void resetNeedsDisplay();

    void setNeedsDisplayOnBoundsChange(bool needsDisplay) { m_needsDisplayOnBoundsChange = needsDisplay; }

    void setOpacity(float opacity) { m_opacity = opacity; setNeedsCommit(); }
    float opacity() const { return m_opacity; }

    void setOpaque(bool opaque) { m_opaque = opaque; setNeedsCommit(); }
    bool opaque() const { return m_opaque; }

    void setPosition(const FloatPoint& position) { m_position = position;  setNeedsCommit(); }
    FloatPoint position() const { return m_position; }

    void setZPosition(float zPosition) { m_zPosition = zPosition; setNeedsCommit(); }
    float zPosition() const {  return m_zPosition; }

    void setSublayerTransform(const TransformationMatrix& transform) { m_sublayerTransform = transform; setNeedsCommit(); }
    const TransformationMatrix& sublayerTransform() const { return m_sublayerTransform; }

    void setTransform(const TransformationMatrix& transform) { m_transform = transform; setNeedsCommit(); }
    const TransformationMatrix& transform() const { return m_transform; }

    bool doubleSided() const { return m_doubleSided; }
    void setDoubleSided(bool doubleSided) { m_doubleSided = doubleSided; setNeedsCommit(); }

    void setPreserves3D(bool preserve3D) { m_preserves3D = preserve3D; }
    bool preserves3D() const { return m_preserves3D; }

    void setIsRootLayer(bool isRootLayer) { m_isRootLayer = isRootLayer; }
    bool isRootLayer() const { return m_isRootLayer; }

    void setIsTransformLayer(const bool isTransformLayer) { m_isTransformLayer = isTransformLayer;}
    const bool isTransformLayer() const { return (m_isTransformLayer); }

    void setOwner(GraphicsLayerWiiU* owner) { m_owner = owner; }

    void setReplicaLayer(LayerWiiU* layer) { m_replicaLayer = layer; }
    LayerWiiU* replicaLayer() { return m_replicaLayer.get(); }

    // These methods typically need to be overwritten by derived classes.
    virtual bool drawsContent() const { return false; }
    virtual void paintContentsIfDirty();

    // FIXME: Create a structure for parameters that are to be inhierated by the children
    // For now we just have the preserve3D and opacity so it not too bad to pass as parameters
    virtual void paint(GraphicsContext *ctx, bool preserve3D, float opacity);

    virtual void paintContents(GraphicsContext *ctx)  {};
	virtual void paintContentLayerIfDirty() {}; 

    virtual void setIsMask(bool) {}

    // These exists just for debugging (via drawDebugBorder()).
    void setBorderColor(const Color&);

    void drawDebugBorder(GraphicsContext *ctx, float width, float height);
    String layerTreeAsText() const;

    void setBorderWidth(float);

    int id() const { return m_layerId; }

    float drawOpacity() const { return m_drawOpacity; }
    void setDrawOpacity(float opacity) { m_drawOpacity = opacity; }

    const TransformationMatrix& drawTransform() const { return m_drawTransform; }
    void setDrawTransform(const TransformationMatrix& matrix) { m_drawTransform = matrix; }


    virtual void cleanupResourcesRecursive();

    float minX() {return m_minX;}
    float maxX() {return m_maxX;}
    float minY() {return m_minY;}
    float maxY() {return m_maxY;}
    float minZ() {return m_minZ;}
    float maxZ() {return m_maxZ;}
    FloatPoint corner1() {return m_corner1;}
    FloatPoint corner2() {return m_corner2;}
    FloatPoint corner3() {return m_corner3;}
    FloatPoint corner4() {return m_corner4;}
    FloatPoint3D normal() {return m_normal;}
    FloatPoint3D origin() {return m_origin;}
    FloatRect boundingBox() {return m_boundingBox;}





protected:
    GraphicsLayerWiiU* m_owner;
    explicit LayerWiiU(GraphicsLayerWiiU* owner);

    // This is called to clean up resources being held in the same context as
    // layerRendererContext(). Subclasses should override this method if they
    // hold context-dependent resources such as textures.
    virtual void cleanupResources();

    void dumpLayer(TextStream&, int indent) const;

    virtual const char* layerTypeAsString() const { return "LayerWiiU"; }
    virtual void dumpLayerProperties(TextStream&, int indent) const;

    FloatRect m_dirtyRect;
    bool m_contentsDirty;

    RefPtr<LayerWiiU> m_maskLayer;

    int m_layerId;

private:

    void paint2D(GraphicsContext *ctx, float opacity);
    void paint3D(GraphicsContext *ctx);
    void process3DTransforms(GraphicsContext *ctx, LayerSorterWiiU *sorter, float opacity);


    void setNeedsCommit();

    void setParent(LayerWiiU*);
    bool hasAncestor(LayerWiiU*) const;

    size_t numChildren() const
    {
        return m_children.size();
    }

    // Returns the index of the child or -1 if not found.
    int indexOfChild(const LayerWiiU*);

    // This should only be called from removeFromParent.
    void removeChild(LayerWiiU*);

    Vector<RefPtr<LayerWiiU> > m_children;
    LayerWiiU* m_parent;

    // Layer properties.
    IntSize m_bounds;

    IntPoint m_scrollPosition;
    FloatPoint m_position;
    FloatPoint m_anchorPoint;
    Color m_backgroundColor;
    Color m_debugBorderColor;
    float m_debugBorderWidth;
    float m_opacity;
    float m_zPosition;
    float m_anchorPointZ;
    bool m_masksToBounds;
    bool m_opaque;
    bool m_needsDisplayOnBoundsChange;
    bool m_doubleSided;
    bool m_isRootLayer;
    bool m_isTransformLayer;
    bool m_preserves3D;

    TransformationMatrix m_transform;
    TransformationMatrix m_sublayerTransform;

    float m_accumulatedOpacity;
    float m_minX;
    float m_maxX;
    float m_minY;
    float m_maxY;
    float m_minZ;
    float m_maxZ;
    FloatPoint m_corner1;
    FloatPoint m_corner2;
    FloatPoint m_corner3;
    FloatPoint m_corner4;
    FloatPoint3D m_normal;
    FloatPoint3D m_origin;
    FloatRect m_boundingBox;

    FloatRect m_frame;

    // Replica layer used for reflections.
    RefPtr<LayerWiiU> m_replicaLayer;

    // Transient properties.
    float m_drawOpacity;
    TransformationMatrix m_drawTransform;

    String m_name;
};


}
#endif // USE(ACCELERATED_COMPOSITING)

#endif
