/*
 * Copyright (C) 2008 Holger Hans Peter Freyther
 *               2008 Kenneth Rohde Christiansen
 *               2009-2010 ProFUSION embedded systems
 *               2009-2010 Samsung Electronics
 *               2012 NINTENDO.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE COMPUTER, INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE COMPUTER, INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "SharedBuffer.h"

#include <wtf/text/CString.h>

#if PLATFORM(WIIU) || PLATFORM(WIIU_DEV)
#include <fcntl.h>
#include <PlatformOS.h>
using namespace LibPlatform;
#else
#include <stdio.h>
#endif
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

namespace WebCore {

PassRefPtr<SharedBuffer> SharedBuffer::createWithContentsOfFile(const String& filePath)
{
#if PLATFORM(WIIU) || PLATFORM(WIIU_DEV)

/*
    CRANK: DLM I changed this to generate something meaningful error messages and to set m_size explicity.
           The version of the code this is based on did not do this, but if this is omitted, the size of the
           SharedBuffer is always zero, which will prevent anything from using the resulting data. Possibly
           the port on which this code was based was broken?
*/

    int fd;
    struct stat fileStat; 
    RefPtr<SharedBuffer> result;

    if (filePath.isEmpty())
        return 0;

    /*
     NSA: O_BINARY is not defined in the open group's unix specification,
     however, the windows platform needs this to correctly display the check
     boxes.
     */
    #if OS(WINDOWS)
    if (!(fd = open(filePath.utf8().data(), O_RDONLY | O_BINARY, 0))) {
    #else
    if (!(fd = open(filePath.utf8().data(), O_RDONLY, 0))) {
    #endif
        OS::Report("SharedBuffer::createWithContentsOfFile: could not open: %s\n", filePath.utf8().data());
        return 0;
    }

    if (fstat(fd, &fileStat)) {
        OS::Report("SharedBuffer::createWithContentsOfFile: could not stat: %s\n", filePath.utf8().data());
        close(fd);
        return 0;
    }

    result = SharedBuffer::create();
    result->m_buffer.resize(fileStat.st_size);
    if (result->m_buffer.size() != fileStat.st_size) {
        close(fd);
        return 0;
    }

    int total_read = 0; 
    while (total_read < fileStat.st_size) {
        int amount_read = read(fd, result->m_buffer.data() + total_read, fileStat.st_size);
        if (amount_read > 0) {
            total_read += amount_read;
        } else {
            break;
        }
    }

    if (total_read < fileStat.st_size) {
        OS::Report("SharedBuffer::createWithContentsOfFile: incomplete read: %s\n", filePath.utf8().data());
        close(fd);
        return 0;
    }

    result->m_size = total_read; 
    close(fd);

    return result.release(); 

#else
    FILE* file;
    struct stat fileStat;
    RefPtr<SharedBuffer> result;

    if (filePath.isEmpty())
        return 0;

    if (!(file = fopen(filePath.utf8().data(), "rb")))
        return 0;

    if (fstat(fileno(file), &fileStat)) {
        fclose(file);
        return 0;
    }

    result = SharedBuffer::create();
    result->m_buffer.resize(fileStat.st_size);
    if (result->m_buffer.size() != fileStat.st_size) {
        fclose(file);
        return 0;
    }

    fread(result->m_buffer.data(), 1, fileStat.st_size, file);
    fclose(file);

    return result.release();
#endif
}

} // namespace WebCore
