/*
 * Copyright (C) 2007 Holger Hans Peter Freyther
 * Copyright (C) 2007, 2008 Alp Toker <alp@atoker.com>
 * Copyright (c) 2012 NINTENDO.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

#ifndef WEB_VIEW_QNX_H
#define WEB_VIEW_QNX_H

#include <wtf/Platform.h> // PLATFORM(...)
#include <wtf/Assertions.h> // ASSERT
#include <wtf/RefPtr.h>
#include <wtf/text/WTFString.h>
#include "PlatformSoftwareKeyboard.h"

#include "IntSize.h"
#include <wk_appinterface.h>
#include <WiiUScrollPoint.h>
#include "WiiUAutoScroller.h"

#if !PLATFORM(WIIU)
#include "WiiURenderParams.h"
#include "renderer/WiiURenderer.h"
#endif

#include <vector>
#include <queue>
#include <map>

#if defined(WIN32) && !PLATFORM(WIIU_DEV)
	#if WEBKIT_IMPLEMENTATION
		#define WEBKIT_API __declspec(dllexport)
	#else//WEBKIT_IMPLEMENTATION
		#define WEBKIT_API __declspec(dllimport)
	#endif//WEBKIT_IMPLEMENTATION
#else//defined(WIN32)
	#define WEBKIT_API __attribute__((visibility("default")))
#endif//defined(WIN32)

namespace WebCore {
    class EditorClientWiiU;
    class HTMLCanvasElement;
    class CanvasRenderingContext;
}

using namespace WTF; 

namespace WebCore {
    class Cursor;
	class Page;
    class FocusController;
	class Frame;
    class GraphicsLayer;
	class WiiUEventManager;
	class PlatformMouseEvent;
	class PlatformTouchEvent;
	class PlatformKeyboardEvent;
	class WiiUDataObject;
	class Node;
	class IntPoint;
	class IntSize;
	class IntRect;
	class PopupMenuWiiU;
	class HTMLElement;
    class Element; 
}

namespace WTF {
class StatisticsWiiU;
}

class WebViewPrivate;

struct PlatformIntRect {
    int x, y, width, height;
};

typedef const struct OpaqueJSContext* JSContextRef;
typedef struct OpaqueJSContext* JSGlobalContextRef;

class WebView;

struct WebViewCallbacks {
    void (*status)(int type, const char *data);
    void (*error)(const char *url, const char *error);
    void (*progress)(int percent);
    void (*paint)(void *texture, WebView *webview);
    WebCore::Page * (*open_window)(const char *url);
};

class WebView {

public:
    enum DisplayType {TV_DISPLAY_TYPE, GAMEPAD_DISPLAY_TYPE, TV_UPPER_DISPLAY_TYPE, TV_LOWER_DISPLAY_TYPE, NO_DISPLAY_TYPE};

    WEBKIT_API WebView(int w, int h,
					   WebViewCallbacks_t *cbs, WebViewPaintCB_t paintCB, WebViewSettings_t *settings,
					   void *surface, bool noSetup, void *userData, WebView* inspectorview, bool inactiveDueToExtDisplay = false);
    virtual ~WebView();

	// callbacks
	void status(int type, const char *data);
	void error(const char *url, const char *error);
	void progress(int percent);
	WebCore::Page *open_window(const char *url);
	bool authenticate(const char *url, char **user, char **pass);

    WEBKIT_API void paint();
	void initSettings(WebViewSettings_t *settings);

    void console_message(const char *message, int line_number, const char *sourceID);
    void js_alert(const char *msg, const wchar_t * msg);
    bool js_confirm(const char * msg);
    bool js_prompt(const char *message, const char *default_result, char **result);

	bool					m_touchScrollEnabled;

    bool					m_touchScrollStart;
    WiiUScrollPoint		m_touchPoint;
	WebCore::Node*			m_lastNodeTouched;
	WebViewScrollStyle		m_scrollStyle;
	WiiUAutoScroller*		m_autoScroller;

	// main API
    WEBKIT_API void goForward();
    WEBKIT_API void goBack();
    WEBKIT_API void reload();
    WEBKIT_API void stop();
    WEBKIT_API void load(const char * url);
	WEBKIT_API void clear();
	WEBKIT_API void pauseMedia();
	WEBKIT_API void resumeMedia();
    
    WEBKIT_API void quitMedia();
    WEBKIT_API void advanceMedia(float time);
    WEBKIT_API void currentTimeMedia(float time);
    WEBKIT_API void startSeekMedia();
    WEBKIT_API void updateSeekMedia(bool isForward);
    WEBKIT_API void endSeekMedia();
    WEBKIT_API void playbackRateMedia(float rate);   
    WEBKIT_API void playPauseMedia(); 

	WEBKIT_API bool isFullScreen() const;
    WEBKIT_API void setFullScreenElement(WebCore::Element *);
    WEBKIT_API void onMousePressed(const WebCore::PlatformMouseEvent & ev);
    WEBKIT_API void onMouseReleased(const WebCore::PlatformMouseEvent & ev);
    WEBKIT_API void onMouseMoved(const WebCore::PlatformMouseEvent & ev);

    WEBKIT_API WebCore::Node* getNodeFromPoint(int x, int y);
    WEBKIT_API bool draggableFromPoint(int x, int y);
    WEBKIT_API bool clickableFromPoint(int x, int y);
    WEBKIT_API bool scrollbarAtPoint(int x, int y);

    WEBKIT_API void setTouchScrollEnabled(bool enabled);
    WEBKIT_API bool getTouchScrollEnabled();
    WEBKIT_API void setScrollbarsEnabled(bool enabled);
    WEBKIT_API bool getScrollbarsEnabled();
    WEBKIT_API void setEmulatedTouchEnabled(bool enabled);
    WEBKIT_API bool getEmulatedTouchEnabled();
    #if ENABLE(TOUCH_EVENTS)
        WEBKIT_API void onTouchPressed(const WebCore::PlatformTouchEvent & ev);
    	WEBKIT_API void onTouchReleased(const WebCore::PlatformTouchEvent & ev);
    	WEBKIT_API void onTouchMoved(const WebCore::PlatformTouchEvent & ev);
    #endif

    WEBKIT_API void setStartVideoInFullscreenEnabled(bool enabled);
    WEBKIT_API bool getStartVideoInFullscreenEnabled();

    WEBKIT_API void onKeyDown(const WebCore::PlatformKeyboardEvent & ev);
    WEBKIT_API void onKeyUp(const WebCore::PlatformKeyboardEvent & ev);
    WEBKIT_API void onSetText(const std::wstring &text);

    WEBKIT_API void scrollLeft(float scale = 1.0f);
    WEBKIT_API void scrollRight(float scale = 1.0f);
    WEBKIT_API void scrollUp(float scale = 1.0f);
    WEBKIT_API void scrollDown(float scale = 1.0f);
    WEBKIT_API void scroll(int dx, int dy);
    WEBKIT_API void scrollNode(WebCore::Node* node, int dx, int dy);

    WEBKIT_API void focusLost();
    WEBKIT_API void focusGained();

    WEBKIT_API void redraw();

    WEBKIT_API void saveScreenshot(const char *filename, bool extendedSaveGamePad = false);
	WEBKIT_API void *captureScreen(int dstWidth, int dstHeight, int srcX = 0, int srcY = 0, int srcWidth = -1, int srcHeight = -1);

	WEBKIT_API void SetUserAgent(const char* agent);
	WEBKIT_API String GetUserAgent();
	WEBKIT_API void SetCustomHeader(const char* header);
	WEBKIT_API std::vector<String> GetCustomHeaders();

#if ENABLE(REQUEST_ANIMATION_FRAME)
    /**
        Allows ChromeClient to notify the WebView that requestAnimationFrame has been called. 
    */
    WEBKIT_API void scheduleAnimation();

    /**
        Allows browser to notify the WebView that animation frame callbacks should be run.
    */
    WEBKIT_API void runScheduledAnimations();
#endif

    // CRANK-ME: Drag and Drop Support
	void startDragging(void * image, RefPtr<WebCore::WiiUDataObject> dataObj);


    void setCursor(const WebCore::Cursor *cursor);

    int x() const { return m_x; }
    int y() const { return m_y; }
    int width() const { return m_width; }
    int height() const { return m_height; }
    
    bool isVisible() const { return m_visible; }    
    void setVisible(bool b);

    void softwareKeyboardInvoke(WebCore::FocusController *pController, SoftwareKeyboard::Mode mode = SoftwareKeyboard::MODE_FULL);
    void softwareKeyboardDismiss();
	void softwareKeyboardEnable(bool bEnable);
	bool softwareKeyboardIsEnabled();

    void setRedrawPending();
    bool redrawPending();
    void setDirtyRect(const PlatformIntRect &dirtyRect);
    void getDirtyRect(PlatformIntRect &dirtyRect);
    bool flushPending();
    void setFlushPending();
	void dumpInnerText();

	static void sendEvent(WiiUBaseEventData* data);
	void enableWebInspector(bool enable);

	WebCore::Page *getPage();
	bool dontQuit();

	//APW: Adding access to global context
	JSGlobalContextRef getJSGlobalContext();
	void *getGraphicsContext(void);
	JSGlobalContextRef getStoredGlobalContext();
	static WebView *GetCurrentWebView(JSContextRef context);


    void setPopup(WebCore::PopupMenuWiiU *popup);
    void setPopupRect(const WebCore::IntRect & rect);
	void SetRenderToTV(bool render_to_tv);
	void SetRenderToGamePad(bool render_to_gamepad, int index);
	bool GetRenderToTV(void) const;
	bool GetRenderToGamePad(void) const;
	int GetGamePadIndex(void) const;

    void collectGarbageNow();
    void collectGarbageSoon();
	void unloadAll();
	void unload(WebCore::Frame *frameP);
    void clearCaches();
    void setMaximumPagesInCache(long pages);
    int maximumPagesInCache();
    void setObjectCacheCapacities(long cacheMinDeadCapacity, long cacheMaxDead, long totalCapacity);
	
    void setCookieJarFilename(char *filename);
    void clearSessionCookies();
    void clearAllCookies();
    void flushCookiesToDisk();
	
	// RJD Crank: Clear the statistics class of all current statistics.  
	void clearStatistics(); 
	
	// RJD Crank: Print out of all the current statistic values.  
	void dumpStatistics(); 
	
	// RJD Crank: Grabs a pointer the statistics instance. 
	StatisticsWiiU *getStatistics(); 

	static void setProxyInfo(const char *host, unsigned long port, WebKitProxyType type, const char *username, const char *password);
	static void setCertificate(const char *cert);
	static void setClientCertificate(const char *cert);
	DisplayType GetDisplayType(void) const;
	void        SetDisplayType(DisplayType type);

	WebViewEventState_t GetEventState();
	void SetEventState(WebViewEventState_t eventState);

	#if !PLATFORM(WIIU)
	void _ForcePaintCall();
	//TODO -- ROB: OpenGL Stuff
	//tWiiURenderParams m_RenderParams;
	WiiURenderer * m_pNewRenderer;
	WiiUWindowType temp_windowType;
	#endif

    void setRootGraphicsLayer(WebCore::GraphicsLayer *);
    void setRootLayerNeedsDisplay();

	void loadExtensionRPLList(WebCore::Frame *frameP=NULL);

	bool m_beenLoaded;
	bool getHasBeenLoaded();

	void blur();
	CString getURLCharacters();
	void sendBGNotification(); 
	void sendFGNotification(); 
	void sendShutdownNotification(); 

	bool redrawEnable() const				{ return m_bRedrawEnable; }
	void setRedrawEnable(bool bEnable)		{ m_bRedrawEnable = bEnable; }

    WebCore::CanvasRenderingContext *getCanvasContext() { return m_pCanvasContext; }

	#if PLATFORM(WIIU)
	struct ScreenshotArg
	{
		std::string filename;
		unsigned char *data;
		int height;
		int width;
		int pitch;
	};

	// Helper function that gets the screenshot buffer
	// 
	ScreenshotArg *_getSaveScreenshotData(const char *filename, bool extendedSaveGamePad);
	void _deleteScreenshotArg(ScreenshotArg *arg);
	#endif

private:
	int						m_x;
	int						m_y;
	int						m_width;
	int						m_height;
	
	bool					m_visible;
	bool                    m_dragActivated;

    bool					m_redrawPending;
    bool                    m_flushPending;
    bool					m_dont_quit;
    PlatformIntRect			m_dirtyRect;
	
    void *					m_userData;

	WebViewPrivate *		m_priv;
	WebCore::Frame* 		m_frame;
	WebCore::Page*			m_page;

	bool 					m_renderToTV;
	bool					m_renderToGamePad;
	bool					m_loadComplete; 
	int						m_gamePadIndex;

	WebViewCallbacks_t *	m_callbacks;
	WebViewPaintCB_t	  (*m_paintCB);
	WebViewSettings_t *		m_settings;
	DisplayType             m_displayType;

    RefPtr<WebCore::HTMLCanvasElement>  m_pCanvasElement;
    WebCore::CanvasRenderingContext *  m_pCanvasContext;

	WebViewEventState_t		m_eventState;

	bool                    m_doingDragAndDrop;

	RefPtr<WebCore::WiiUDataObject>	m_currentDragDataObject;

	bool					m_bRedrawEnable;

	bool 					m_startVideoInFullscreen;

	typedef std::vector<WebCore::HTMLElement*> ElementList;
	ElementList m_pausedMediaElements;

	inline void PaintCall(void);
};

class WebViewManager {
public:


	WebViewManager(void);
	~WebViewManager(void);

	void CreateWebViews(WebViewCallbacks_t *callbacks, WebViewSettings_t *settings, unsigned int gamepad_web_view_count, WebViewInspectorLocation_t location);
	void DestroyWebViews(bool bOnWebkitThread);

	WebView *GetTVWebView(void);
	WebView *GetInspectorWebView(void);
	WebView *GetGamePadWebView(unsigned int index);

	WebView::DisplayType GetDisplayType(WebView *web_view);

	int GetNumWebViews(void);
	int GetNumGamePadWebViews(void);
	
	WebViewCallbacks_t *getCallbacks(); 
	WebViewSettings_t *getSettings(); 
	WebViewInspectorLocation_t getLocation(); 
	
	void setRestartWebViews(bool restart, float fadeOutTimeMS = 0, const char * vConfigXML = 0, const unsigned int vConfigXMLLength = 0);
	const char * getConfigXMLData();
	const unsigned int getConfigXMLDataLength();
	bool restartWebViews();
	float getRestartWebViewFadeOutTime(void);

	void blockWebViewRestart(bool shouldBlock, int blockSource);

	static void setScreenshotCompleteCallback(void (*fn)(const char *, bool));
	static void setSendInspectorConnectedCallback(void (*fn)());
	static void callScreenshotComplete(const char *filename, bool succeeded);
	static void callSendInspectorServerStarted();

	//After adding callSendInspectorServerStarted() which follows the design pattern set forth by other
	//functions such as callScreenshotComplete(), where several functions are required to get a message
	//across to the dashboard, I'm adding this so that we can stop adding those kinds of functions.
	static void setSendDashboardSpecificMessageCallback(void(*fn)(unsigned int));
	static void callSendDashboardSpecificMessage(unsigned int vMessage);

	#if PLATFORM(WIIU)
	static void saveScreenshot(WebView *webview, const char *filename, bool extendedSaveGamePad);
	static void abortScreenshots(); // Blocks until all screenshot threads have ended
	#endif

	static void addScreenshotCompleteCallback(void (*fn)(const char *, bool));
	static void removeScreenshotCompleteCallback(void (*fn)(const char *, bool));

	// screenshotProgress(filename, data, dataLen)
	static void setScreenshotProgressCallback(bool (*fn)(const char *, void *, int), int progressFrequency);

	// Pass the previous return value from this for lastRet
	// Returns -1 when the screenshot has failed and should be aborted
	// Returns 0 when the progress callback is called and data can be safely cleared
	// If the return is a positive value, then data should continue to point to somewhere in the same contiguous buffer
	static int callScreenshotProgress(const char *filename, void *data, int dataLen, bool final, int lastRet = 0);
	
	void SetPanicExit(void);

private:
	WebView *				m_TVwebView;
	std::vector<WebView *>	m_GamePadWebViews;
	pthread_mutex_t			m_GamePadWebViewsMutex;
	WebView *               m_inspectorView;
	WebViewCallbacks_t *    m_callbacks;
	WebViewSettings_t *		m_settings;
	WebViewInspectorLocation_t 	m_location;
	bool					m_bPanicExit;

    
	//Restart support
	bool m_restartWebViews;
	float m_fadeOutTimeMS;
	char * m_pNewConfigXMLData;
	unsigned int m_NewConfigXMLLength;
	int m_blockRestart;
	bool m_hasBlockedRestart;
	char *m_pQueuedRestartData;
	unsigned int m_QueuedRestartDataLength;

	#if PLATFORM(WIIU)
	// Screenshot queuing
	struct WebViewScreenshotThread {
		typedef std::queue<WebView::ScreenshotArg*> ScreenshotQueue;
		ScreenshotQueue m_queuedScreenshots;
		pthread_mutex_t m_mutex;
		pthread_t m_thread;
		bool m_threadStopped;
		WebView *m_webview;
		long long m_lastScreenshotTime;
	};
	typedef std::map<WebView*, WebViewScreenshotThread> ActiveScreenshotThreads;
	static ActiveScreenshotThreads *m_screenshots;
	#endif

	static void *screenshotThreadFunc(void *arg);

	static void (*m_SendInspectorServerStartedCallback)();
	static void (*m_SendDashboardSpecificMessageCallback)(unsigned int);

	// Screenshot complete support
	typedef void (*ScreenshotCompleteCallback)(const char *, bool);
	static std::vector<ScreenshotCompleteCallback> *m_screenshotCompleteCallbacks;

	// Screenshot progress support
	static bool (*m_screenshotProgressCallback)(const char *, void *, int);
	static int m_screenshotProgressFrequency;
};

#endif
